/**
 * The <code>FasterTermStructure</code> class is a faster
 * version of the AbstractTermStructure class.  
 *
 * Now it is just a copy of the AbstractTermStructure class.
 * You should modify the code to use the faster compound state
 * prices method and your YieldVolObject class.
 *
 * You will need to replace the methods that have "slow" in
 * their title with their faster equivalents.
 *
 * Don't forget to put your name in the author field below!
 *
 * @author  
 */

package interestRate;

import mathUtil.*;

public abstract class FasterTermStructure {


  private static final boolean DEBUG=false;

  //time horizon
  int T;

  //Used by the Newton-Raphson solver
  SlowYieldVolObject slowYieldVolObj;
  NewtonRaphson slowSolver;

  //parameters of the BDT model
  double r[];
  double k[];

  //bond yields and yield volatilities
  //at time 0
  double yield[];
  double volatilities[];

  //nodes[i] points to link list of
  //nodes with time i
  LinkList nodes[];


  /**
    Constructor takes following arguments.
    @params T Time Horizon
    @params yield Array of Bond yields
    @params volatilites Array of Volatilities
    */
  public FasterTermStructure(int T, double yield[],
			       double volatilities[]){
			
      //set the time horizon
      this.T = T;

      //allocate space for the parameters
      r = new double[T+1];
      k = new double[T+1];

      //copy the bond yields and yield volatilities
      int arrayLength = yield.length;
      
      //assume volatility array has the same length
      //as yield array
      this.yield = new double[arrayLength];
      this.volatilities = new double[arrayLength];

      //copy the arrays 
      System.arraycopy(yield,0,this.yield,0,arrayLength);
      System.arraycopy(volatilities,0,this.volatilities,0,arrayLength);
    
  
      //allocate the linked-list
      nodes = new LinkList[T+1];
      for(int i=0; i <= T; i++)
	   nodes[i] = new LinkList();

      //Get the yield-vol object
      slowYieldVolObj = new SlowYieldVolObject(this);

      //Instantiate the solver
      slowSolver = new NewtonRaphson(slowYieldVolObj);

  }//end of FasterTermStructure
    

  /**
     The form of the short rate at the node (time,up_ticks)
     */
  public abstract double F(int time, int up_ticks,
			double r, double k);


  /**
     Generate the entire pdag upto time horizon T.
     */
  public void GenPdag() {


    if (DEBUG) System.out.println("Entered GenPdag");
    nodes[0].Insert(new Key(0,0));


    for(int t=0; t < T; t++) {
      Node x=nodes[t].head;

      
      while (x != null) {
	if (DEBUG) {
	  System.out.println("Considering node: ");
	  x.key.print();
	}

	int up_ticks = ((Key)(x.key)).up_ticks;


	//Generate successor nodes
	Key up_key = new Key(t+1,up_ticks+1);
	Key down_key = new Key(t+1,up_ticks);
	x.succ[0] = nodes[t+1].Insert(up_key);
	x.succ[1] = nodes[t+1].Insert(down_key);
	x = x.next;
      }//end of while
    }//end of for

	
    if (DEBUG) System.out.println("Leaving GenPdag");
      

  }//end of GenPdag 


  /**
     print the entire dag.
     */
  public void print() {


    try {
      for(int t=0; t <= T; t++) {
	System.out.print("Nodes at time ");
	System.out.println(t);
	System.out.println("------------BEGIN-----------------");
	nodes[t].print();
	System.out.println("------------END------------------");
      }
    }//
    catch (LinkListException e) {
      System.err.println("Shouldn't happen! "+e.getMessage());
      System.exit(1);
    }
  }//end of print

  //price of the bond  of a given maturity
  //(t,up_ticks). Assume that the parameters
  //r, k are known upto time maturity-1
  private double slowPrice(int t, int up_ticks,
	       int maturity) {

    //Handle the base case
    if (t == maturity) return(1);
    else {
      //Recursive call price on successor node
      double price_up = slowPrice(t+1,up_ticks+1,maturity);
      double price_down = slowPrice(t+1,up_ticks,maturity);
      
      double returnPrice = 
	(0.5/(1+F(t,up_ticks,r[t],k[t])))*(price_up+price_down);

      if (DEBUG) {
	System.out.print("slowPrice:AbstractTermStructure returnPrice ");
	System.out.println(returnPrice);
      }
      return(returnPrice);
    }//end of else

	 
    

  }//end of slowPrice


  /**
     yield of the bond of a given maturity 
     at the node (t,up_ticks).
     */
  public double slowYield(int t, int up_ticks,
	       int maturity) {

    double bondPrice = slowPrice(t,up_ticks,maturity);

    double bondYield = Math.pow(1.00/bondPrice,1.00/(maturity-t))-1;
    if (DEBUG) {
	System.out.print("slowYield:AbstractTermStructure bondYield ");
	System.out.println(bondYield);
    }
      
    return(bondYield);

  }//end of slowYield


  /**
     volatality of the yield of the bond 
     at node (t,up_ticks)
     */
  double slowLogVol(int t, int up_ticks,
	     int maturity) {

    double up_yield = Math.log(slowYield(t+1,up_ticks+1,maturity));
    double down_yield = Math.log(slowYield(t+1,up_ticks,maturity));

    double expectedSquareVal = 0.5*(up_yield*up_yield+
				    down_yield*down_yield);

    double expectedVal = 0.5*(up_yield+down_yield);

    return(Math.sqrt(expectedSquareVal-expectedVal*expectedVal));


  }//end of slowLogVol



  //solve for the parameters at time t
  private void  slowSolve(int t) {

    //Handle the base case
    if (t==0) {
      r[0] = yield[0];
      k[0] = 1;
    }
    else {
      //update the maturity in the slowYieldVolObj
      slowYieldVolObj.maturity = t+1;
      double initialVal[] = new double[2];

      //Set the initial value of the parameters
      //time t to parameter values at time t-1
      initialVal[0] = r[t-1];
      initialVal[1] = k[t-1];
      double result[] = slowSolver.solve(initialVal);
      r[t] = result[0];
      k[t] = result[1];

    }//end of else

  }//end of slowSolve


  /**
     Solve for the entire interest rate tree.
     */
  public void slowSolve() {
    //Call slowSolve iteratively
    for (int t=0; t <= T; t++) {
      slowSolve(t);
      //fill the short-rate at the nodes
      Node x=nodes[t].head;
      while (x != null) {
	Key key = (Key)x.key;
	key.short_rate = F(t,key.up_ticks,r[t],k[t]);
	x=x.next;
      }
    }
  }//end of slowSolve

      
}//end of FasterTermStructure
  
