package mathUtil;

/** Has all the Matrix utilities built it.
    @author Somesh Jha
    */

public class Matrix {

  public int dimension;
  public double data[][];
  private static boolean DEBUG = false;

  /** The constructor takes an input parameter
      and the array to instantiate the matrix 
  */

  public Matrix(int n,
	        double input[][]) {
    dimension = n;
    data = input;
  }//end of constructor

  /** Print the matrix in row major order.
     
    */
  public void print(){
    System.out.println("The matrix is: ------");
    for(int i=0; i < dimension; i++) {
      System.out.print("Row ");
      System.out.print(i);
      System.out.print(" ");
      for(int j=0; j < dimension; j++) {
	System.out.print(data[i][j]);
	System.out.print(" ");
      }
      System.out.println();
    }
    System.out.println("---------------------");
  }//end of print

  /** Get the slice of the matrix between i and
      j. i should be less than j **/
  private Matrix slice(int i, int j) {
    int n= j-i+1;
    double input[][] = new double[n][n];

    for (int k= i; k <= j; k++) 
      for(int m=i; m <= j; m++) {
	input[k-i][m-i] = this.data[k][m];
      }
    
    return (new Matrix(n,input));
  }//end of slice


  /** Multiple this matrix with a vector on the right.
   */

  public double[] multiplyVector(double values[]) {

    double returnVal[] = new double[dimension];
    for(int i=0; i < dimension; i++) {
      returnVal[i]=0.0;
      for(int j=0; j < dimension; j++)
	returnVal[i] += data[i][j]*values[j];
    }

    return(returnVal);
  }//end of multiplyVector


  /** Multiply this matrix on the right with another matrix
    */
  public Matrix multiplyRight(Matrix second) {
    if (this.dimension != second.dimension) return(null);

    int n = second.dimension;
    double input[][] = new double[n][n];

    for(int i=0; i < n; i++)
      for(int j=0; j < n; j++) {
	input[i][j]=0;
	for(int k=0; k < n; k++) 
	  input[i][j] += this.data[i][k]*second.data[k][j];
      }

    return (new Matrix(n,input));
  }//end of multiply



  /** Multiply this matrix on the left with another matrix
    */
  public Matrix multiplyLeft(Matrix second) {
    if (this.dimension != second.dimension) return(null);

    int n = second.dimension;
    double input[][] = new double[n][n];

    for(int i=0; i < n; i++)
      for(int j=0; j < n; j++) {
	input[i][j]=0;
	for(int k=0; k < n; k++) 
	  input[i][j] += second.data[i][k]*data[k][j];
      }

    return (new Matrix(n,input));
  }//end of multiplyLeft


  //Find an r such that data[r][0] not equal to 0
  //return -1 if such element is not found
  private int findPivot() {
    for(int i=0; i < dimension ; i++) 
      if (data[i][0] != 0) return(i);
    

    return(-1);
  }//end of findPivot

  //Return a permutation matrix that exchanges 
  //rows i and j
  private Matrix permutationMatrix(int i,
				   int j) {
    double input[][] = new double[dimension][dimension];
    
    for(int k=0; k < dimension; k++) 
      for(int m=0; m < dimension; m++) {
	if (k == i && m== j) input[k][m] = 1;
	else if (k == j && m == i) input[k][m] = 1;
	else if (k == m) input[k][m] = 1;
	else input[k][m] = 0;
      }

    return (new Matrix(dimension,input));
  }//end of permutationMatrix


  //Return the matrix to multiply by
  private Matrix rightMatrix() {
	double input[][] = new double[dimension][dimension];

	input[0][0] = 1/data[0][0];
	for(int i=1; i < dimension; i++) 
	  input[i][0] = -data[i][0]/data[0][0];
	
	for(int i=1; i < dimension; i++)
	  for(int j=0; j < dimension; j++) 
	    if (i == j) input[j][i] = 1;
	    else input[j][i] = 0;
	
	return (new Matrix(dimension,input));
  }//end of rightMatrix


  private Matrix rightMatrixLowerTriangular() {
    

    double input[][] = new double[dimension][dimension];

	input[dimension-1][dimension-1] = 1;
	for(int i=0; i < dimension-1; i++) 
	  input[i][dimension-1] = -data[i][dimension-1];
	
	for(int i=0; i < dimension-1; i++)
	  for(int j=0; j < dimension; j++) 
	    if (i == j) input[j][i] = 1;
	    else input[j][i] = 0;
	
	return (new Matrix(dimension,input));
  }//end of rightMatrixLowerTriangular


  /** Assume that this matrix is lower triangular and
      invert it. The algorithm is recursive.
      */
  public Matrix invertMatrixLowerTriangular() {

    //If the dimension is 1, have to do nothing
    if (dimension == 1) return(this);

    //Get the matrix that makes all entries (but the
    //last one) 0
    Matrix m = rightMatrixLowerTriangular();
    
    //Multiple the current matrix with m
    Matrix a = multiplyLeft(m);

    //Get the slice of the matrix that omits
    //the last row and column
    Matrix aSlice = a.slice(0,dimension-2);

    //Invert the sliced matrix
    Matrix m1 = aSlice.invertMatrixLowerTriangular();

    //Pad the matrix to increase the dimesnion
    double input[][] = new double[dimension][dimension];
    for(int i=0; i < dimension-1; i++) 
      for(int j=0; j < dimension-1; j++)
	input[i][j] = m1.data[i][j];
    
    for(int i=0; i < dimension; i++) {
      input[dimension-1][i]=0;
      input[i][dimension-1]=0;
    }
    input[dimension-1][dimension-1] =1;

    Matrix m2= new Matrix(dimension,input);
    
    return (m2.multiplyRight(m));

  }//end of invertMatrixLowerTraingular



  /** This routine converts a given a matrix into an
      upper triangular matrix, i.e., all elements below
      the diagonal are zero and all elements on the diagonal
      are one. The routine is recursive. Returns null if the
      matrix is singular.
      */

  private Matrix convertLowerTriangular() {
    
    //If dimension is one, return the reciprocal
    //of the element.

    if (dimension == 1) {
      if (data[0][0] == 0.0) return(null);

      double input[][] = new double[1][1];
      input[0][0] = 1/data[0][0];

      return (new Matrix(1,input));
    }


    //Find a pivot element that is non-zero
    Matrix m;
    int pivot = findPivot();
    if (pivot == -1) return(null);
    if (pivot == 0) {
      m = rightMatrix();
    }
    else {
      //exchange the 0-th and the pivot row
      Matrix perm = permutationMatrix(0,pivot);
      m = perm.multiplyLeft(rightMatrix());
    }

    //zap the first column
    Matrix a = multiplyLeft(m);
    if (DEBUG) {
      System.out.println("convertLowerTriangular:a");
      a.print();
    }


    //slice the first column
    Matrix aSlice = a.slice(1,dimension-1);
    if (DEBUG) {
      System.out.println("convertLowerTriangular:aSlice");
      aSlice.print();
    }

    Matrix m1 = aSlice.convertLowerTriangular();
    //return null if the rest of the matrix is 
    //singular
    if (m1 == (Matrix)null) return(null);


    double input[][] = new double[dimension][dimension];
    for (int i=1; i < dimension; i++)
      for(int j=1; j < dimension; j++) 
	input[i][j] = m1.data[i-1][j-1];

    for(int i=0; i < dimension; i++) {
      input[0][i] =0;
      input[i][0]=0;
    }

    input[0][0]=1;
    Matrix m2= new Matrix(dimension,input);
    return(m2.multiplyRight(m));

  }//end of convertLowerTriangular


  /** Return the inverse of the matrix. Return null
      if the matrix is singular.
      */
  public Matrix invertMatrix() {
    //convert the matrix to lower triangular
    Matrix m = convertLowerTriangular();
    if (m == null) return(null);

    //invert the lower triangualr matrix
    Matrix a = multiplyLeft(m);
    Matrix m1= a.invertMatrixLowerTriangular();
    return(m1.multiplyRight(m));
  }//end of invertMatrix


    
    


}//end of class Matrix
