
package mathUtil;

/** Implements a Newton Raphson Solver
    @author Somesh Jha
    */
public class NewtonRaphson {
  int noOfVars;
  AbstractFunctionObject funcObj;
  private static boolean DEBUG=false;
  private static double DELTA = 0.00001;
  private static double EPSILON=0.00001;
  private static double MAX=10;

  /** Solver takes argument as a function object.
    */
  public NewtonRaphson(AbstractFunctionObject funcObj) {
    noOfVars = funcObj.n;
    this.funcObj = funcObj;
  }//end of NewtonRaphson


  private double[] evaluate(double values[]) {
    double functionValues[] = new double[noOfVars];

    for(int i=0; i < noOfVars; i++) 
      functionValues[i] = funcObj.evaluate(i,values);
    
    return(functionValues);
  }//end of evaluate

  private double norm(double functionVals[]) {
    
    double returnVal = 0.0;
    for(int i=0; i < noOfVars; i++) 
      returnVal += functionVals[i]*functionVals[i];

    return(returnVal);
  }//end of norm

  private Matrix  getJacobian(double values[]) {
    double input[][] = new double[noOfVars][noOfVars];
    
    for(int i=0; i < noOfVars; i++) 
      for(int j=0; j < noOfVars; j++)
	input[i][j] = funcObj.derivative(i,j,values,DELTA);

    return(new Matrix(noOfVars,input));
  }//end of getJacobian
    

  /** Provided the initial seed solve the
      system of equations.
      */
  public double[] solve(double initialVal[]) {
    double val[]=initialVal;
    
    double functionVal[] = evaluate(val);
    if (DEBUG) {
      System.out.print("NewtonRaphson:solve ");
      for(int i=0; i < noOfVars; i++)
	System.out.println(functionVal[i]);
    }
    
    int counter=1;
    while (norm(functionVal) > EPSILON) {
      Matrix jacobian = getJacobian(val);
      if (DEBUG) jacobian.print();

      Matrix inverseJacobian = jacobian.invertMatrix();
      if (DEBUG) inverseJacobian.print();

      if (DEBUG) {
	Matrix tempMatrix = jacobian.multiplyLeft(inverseJacobian);
	tempMatrix.print();
      }

      double newVal[] = inverseJacobian.multiplyVector(functionVal);
      for(int i=0; i < noOfVars; i++) 
	val[i] = val[i]-newVal[i];
      functionVal = evaluate(val);

      if (DEBUG) {
	System.out.print("NewtonRaphson:solve ");
	for(int i=0; i < noOfVars; i++)
	  System.out.println(functionVal[i]);
      }


      counter++;
      if (counter > MAX) break;
    }

    return(val);

  }//end of solve
      
}//end of class NewtonRaphson      

  
